/*
 * Copyright 2017-present Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.onosproject.nemo.model.object.match.item.definitions;

import static com.google.common.base.MoreObjects.toStringHelper;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import org.onosproject.nemo.model.common.MatchItemName;

import com.google.common.collect.Maps;

/**
 * The builder of match item definition.
 */
public class MatchItemDefinitionBuilder {

    private Map<Class<? extends MatchItemDefinition>, MatchItemDefinition>
        augmentation = Maps.newHashMap();
    private MatchItemDefinitionKey key;
    private MatchItemName matchItemName;
    private MatchItemDefinition.MatchItemValueType matchItemValueType;

    /**
     * Creates a default match item definition builder.
     */
    public MatchItemDefinitionBuilder() {
    }

    /**
     * Creates an match item definition builder from a given match item
     *  definition.
     *
     * @param base the match item definition
     */
    public MatchItemDefinitionBuilder(MatchItemDefinition base) {
        if (base.getKey() == null) {
            key = new MatchItemDefinitionKey(base.getMatchItemName());
            matchItemName = base.getMatchItemName();
        } else {
            key = base.getKey();
            matchItemName = key.getMatchItemName();
        }
        matchItemValueType = base.getMatchItemValueType();
        MatchItemDefinitionImpl impl = (MatchItemDefinitionImpl) base;
        if (!impl.augmentation.isEmpty()) {
            augmentation = new HashMap<>(impl.augmentation);
        }
    }

    /**
     * Sets the match item definition key to be used by the builder.
     *
     * @param value the match item definition key value
     * @return self
     */
    public MatchItemDefinitionBuilder Key(MatchItemDefinitionKey value) {
        key = value;
        return this;
    }

    /**
     * Sets the match item name to be used by the builder.
     *
     * @param value the match item value
     * @return self
     */
    public MatchItemDefinitionBuilder MatchItemName(MatchItemName value) {
        matchItemName = value;
        return this;
    }

    /**
     * Sets the match item value type to be used by the builder.
     *
     * @param value match item definition value type
     * @return self
     */
    public MatchItemDefinitionBuilder MatchItemValueType(MatchItemDefinition.MatchItemValueType value) {
        matchItemValueType = value;
        return this;
    }

    /**
     * Adds an augmentation to the augmentation list. If the given augmentation
     * is null, then removes the augmentation type from the map.
     *
     * @param augmentationType the type of argumentation
     * @param augmentation the augmentation
     * @return self
     */
    public MatchItemDefinitionBuilder addAugmentation(
            Class<? extends MatchItemDefinition> augmentationType,
            MatchItemDefinition augmentation) {
        if (augmentation == null) {
            return removeAugmentation(augmentationType);
        }
        this.augmentation.put(augmentationType, augmentation);
        return this;
    }

    /**
     * Deletes an augmentation from the augmentation list.
     *
     * @param augmentationType the type of argumentation
     * @return self
     */
    public MatchItemDefinitionBuilder removeAugmentation(
            Class<? extends MatchItemDefinition> augmentationType) {
        augmentation.remove(augmentationType);
        return this;
    }

    /**
     * Builds an match item definition from this builder instance.
     *
     * @return match item definition
     */
    public MatchItemDefinition build() {
        return new MatchItemDefinitionImpl(this);
    }

    /**
     * The implementation of match item definition.
     */
    private static final class MatchItemDefinitionImpl
            implements MatchItemDefinition {

        private final MatchItemDefinitionKey key;
        private final MatchItemName matchItemName;
        private final MatchItemValueType matchItemValueType;
        private final Map<Class<? extends MatchItemDefinition>,
                MatchItemDefinition> augmentation;

        private MatchItemDefinitionImpl(MatchItemDefinitionBuilder base) {
            if (base.key == null) {
                key = new MatchItemDefinitionKey(base.matchItemName);
                matchItemName = base.matchItemName;
            } else {
                key = base.key;
                matchItemName = key.getMatchItemName();
            }
            matchItemValueType = base.matchItemValueType;
            augmentation = new HashMap<>(base.augmentation);
        }

        @Override
        public MatchItemDefinitionKey getKey() {
            return key;
        }

        @Override
        public MatchItemName getMatchItemName() {
            return matchItemName;
        }

        @Override
        public MatchItemValueType getMatchItemValueType() {
            return matchItemValueType;
        }

        @Override
        public int hashCode() {
            return Objects.hash(key, matchItemName, matchItemValueType,
                    augmentation);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            MatchItemDefinition other = (MatchItemDefinition) obj;
            if (!Objects.equals(key, other.getKey())) {
                return false;
            }
            if (!Objects.equals(matchItemName, other.getMatchItemName())) {
                return false;
            }
            if (!Objects.equals(matchItemValueType, other.getMatchItemValueType())) {
                return false;
            }
            MatchItemDefinitionImpl otherImpl = (MatchItemDefinitionImpl) obj;
            return Objects.equals(augmentation, otherImpl.augmentation);
        }

        @Override
        public String toString() {
            return toStringHelper(this)
                    .add("key", key)
                    .add("matchItemName", matchItemName)
                    .add("matchItemValueType", matchItemValueType)
                    .add("augmentation", augmentation)
                    .toString();
        }
    }
}