/*
 * Copyright 2016-present Open Networking Laboratory
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.onosproject.nemo.model.object.node.definitions;

import static com.google.common.base.MoreObjects.toStringHelper;
import com.google.common.collect.Maps;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.onosproject.nemo.model.common.NodeType;
import org.onosproject.nemo.model.object.PropertyDefinitions;
import org.onosproject.nemo.model.object.property.definitions.PropertyDefinition;

/**
 * The builder of node definition.
 */
public class NodeDefinitionBuilder {
    private Map<Class<? extends NodeDefinition>, NodeDefinition>
            augmentation = Maps.newHashMap();
    private NodeDefinitionKey key;
    private NodeType nodeType;
    private List<PropertyDefinition> propDefs;

    /**
     * Creates a default node definition builder.
     */
    public NodeDefinitionBuilder() {
    }

    /**
     * Creates a node definition builder from a given definitions.
     *
     * @param definitions the property definitions
     */
    public NodeDefinitionBuilder(PropertyDefinitions definitions) {
        propDefs = definitions.getPropertyDefinitions();
    }

    /**
     * Creates a node definition builder from a given node definition.
     *
     * @param base the node definition object
     */
    public NodeDefinitionBuilder(NodeDefinition base) {
        if (base.getKey() == null) {
            key = new NodeDefinitionKey(base.getNodeType());
            nodeType = base.getNodeType();
        } else {
            key = base.getKey();
            nodeType = key.getNodeType();
        }
        propDefs = base.getPropertyDefinitions();
    }

    /**
     * Sets the key to be used by the builder.
     *
     * @param key key value
     * @return self
     */
    public NodeDefinitionBuilder key(NodeDefinitionKey key) {
        this.key = key;
        return this;
    }

    /**
     * Sets the node type to be used by the builder.
     *
     * @param type node type
     * @return self
     */
    public NodeDefinitionBuilder nodeType(NodeType type) {
        nodeType = type;
        return this;
    }

    /**
     * Sets the property definitions to be used by the builder.
     *
     * @param definition property definition
     * @return self
     */
    public NodeDefinitionBuilder propertyDefinition(
            List<PropertyDefinition> definition) {
        propDefs = definition;
        return this;
    }

    /**
     * Adds an augmentation to the augmentation list. If the given augmentation
     * is null, then removes the augmentation type from the map.
     *
     * @param augmentationType the type of argumentation
     * @param augmentation the augmentation
     * @return self
     */
    public NodeDefinitionBuilder addAugmentation(
            Class<NodeDefinition> augmentationType,
            NodeDefinition augmentation) {
        if (augmentation == null) {
            return removeAugmentation(augmentationType);
        }
        this.augmentation.put(augmentationType, augmentation);
        return this;
    }

    /**
     * Deletes an augmentation from the augmentation list.
     *
     * @param augmentationType the type of argumentation
     * @return self
     */
    public NodeDefinitionBuilder removeAugmentation(
        Class<NodeDefinition> augmentationType) {
        augmentation.remove(augmentationType);
        return this;
    }

    /**
     * Builds a node definition from this builder instance.
     *
     * @return node definition
     */
    public NodeDefinition build() {
        return new NodeDefinitionImpl(this);
    }

    /**
     * The implementation of node definition interface.
     */
    private static final class NodeDefinitionImpl implements NodeDefinition {

        private final NodeDefinitionKey key;
        private final NodeType nodeType;
        private final List<PropertyDefinition> propertyDefinitions;
        private final Map<Class<? extends NodeDefinition>, NodeDefinition>
                augmentation;

        private NodeDefinitionImpl(NodeDefinitionBuilder base) {
            if (base.key == null) {
                key = new NodeDefinitionKey(base.nodeType);
                nodeType = base.nodeType;
            } else {
                key = base.key;
                nodeType = key.getNodeType();
            }
            propertyDefinitions = base.propDefs;
            augmentation = new HashMap<>(base.augmentation);
        }

        @Override
        public NodeDefinitionKey getKey() {
            return key;
        }

        @Override
        public NodeType getNodeType() {
            return nodeType;
        }

        @Override
        public List<PropertyDefinition> getPropertyDefinitions() {
            return propertyDefinitions;
        }

        @Override
        public int hashCode() {
            return Objects.hash(key, nodeType, propertyDefinitions,
                    augmentation);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            NodeDefinition other = (NodeDefinition) obj;
            if (!Objects.equals(key, other.getKey())) {
                return false;
            }
            if (!Objects.equals(nodeType, other.getNodeType())) {
                return false;
            }
            if (!Objects.equals(propertyDefinitions, other.
                    getPropertyDefinitions())) {
                return false;
            }
            NodeDefinitionImpl otherImpl = (NodeDefinitionImpl) obj;
            return Objects.equals(augmentation, otherImpl.augmentation);
        }

        @Override
        public String toString() {
            return toStringHelper(this)
                    .add("key", key)
                    .add("nodeType", nodeType)
                    .add("propertyDefinition", propertyDefinitions)
                    .add("augmentation", augmentation)
                    .toString();
        }
    }
}