/*
 * Copyright 2016-present Open Networking Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.onosproject.nemo.model.operation.condition.parameter.definitions;

import static com.google.common.base.MoreObjects.toStringHelper;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import org.onosproject.nemo.model.common.ParameterName;
import org.onosproject.nemo.model.operation.condition.parameter.definitions.condition.parameter.definition.ParameterMatchPatterns;

/**
 * Representation of building a condition parameter definition.
 */
public class ConditionParameterDefinitionBuilder {

    private Map<Class<? extends ConditionParameterDefinition>,
            ConditionParameterDefinition> augmentation = new HashMap<>();
    private ConditionParameterDefinitionKey key;
    private ParameterMatchPatterns parameterMatchPatterns;
    private ParameterName parameterName;
    private ConditionParameterDefinition.ParameterValueType parameterValueType;

    /**
     * Creates a default condition parameter definition builder.
     */
    public ConditionParameterDefinitionBuilder() {
    }

    /**
     * Creates a condition parameter builder from a given condition parameter
     * definition.
     *
     * @param base the condition parameter definition
     */
    public ConditionParameterDefinitionBuilder(
            ConditionParameterDefinition base){
        if (base.getKey() == null) {
            key = new ConditionParameterDefinitionKey(base.getParameterName());
            parameterName = base.getParameterName();
        } else {
            key = base.getKey();
            parameterName = key.getParameterName();
        }
        parameterMatchPatterns = base.getParameterMatchPatterns();
        parameterValueType = base.getParameterValueType();
        if (base instanceof ConditionParameterDefinitionImpl) {
            ConditionParameterDefinitionImpl impl =
                    (ConditionParameterDefinitionImpl) base;
            if (!impl.augmentation.isEmpty()) {
                this.augmentation = new HashMap<>(impl.augmentation);
            }
        }
    }

    /**
     * Sets the condition parameter definition key to be used by the builder.
     *
     * @param key condition parameter definition key
     * @return self
     */
    public ConditionParameterDefinitionBuilder key(
            ConditionParameterDefinitionKey key) {
        this.key = key;
        return this;
    }

    /**
     * Sets the condition parameter match patterns to be used by the builder.
     *
     * @param patterns condition parameter match patterns
     * @return self
     */
    public ConditionParameterDefinitionBuilder parameterMatchPatterns(
            ParameterMatchPatterns patterns) {
        parameterMatchPatterns = patterns;
        return this;
    }

    /**
     * Sets the parameter name to be used by the builder.
     *
     * @param name the parameter name
     * @return self
     */
    public ConditionParameterDefinitionBuilder parameterName(
            ParameterName name) {
        parameterName = name;
        return this;
    }

    /**
     * Sets the parameter value type to be used by the builder.
     *
     * @param type parameter value type
     * @return self
     */
    public ConditionParameterDefinitionBuilder parameterValueType(
            ConditionParameterDefinition.ParameterValueType type) {
        parameterValueType = type;
        return this;
    }

    /**
     * Adds an augmentation to the augmentation list. If the given augmentation
     * is null, then removes the augmentation type from the map.
     *
     * @param augmentationType the type of argumentation
     * @param augmentation the augmentation
     * @return self
     */
    public ConditionParameterDefinitionBuilder addAugmentation(
            Class<? extends ConditionParameterDefinition> augmentationType,
            ConditionParameterDefinition augmentation) {
        if (augmentation == null) {
            return removeAugmentation(augmentationType);
        }
        this.augmentation.put(augmentationType, augmentation);
        return this;
    }

    /**
     * Deletes an augmentation from the augmentation list.
     *
     * @param augmentationType the type of argumentation
     * @return self
     */
    public ConditionParameterDefinitionBuilder removeAugmentation(
            Class<? extends ConditionParameterDefinition> augmentationType) {
        augmentation.remove(augmentationType);
        return this;
    }

    /**
     * Builds a condition parameter definition from this builder instance.
     *
     * @return condition parameter definition
     */
    public ConditionParameterDefinition build() {
        return new ConditionParameterDefinitionImpl(this);
    }

    /**
     * The implementation of condition parameter definition.
     */
    private static final class ConditionParameterDefinitionImpl
            implements ConditionParameterDefinition {

        private final ConditionParameterDefinitionKey key;
        private final ParameterMatchPatterns parameterMatchPatterns;
        private final ParameterName parameterName;
        private final ParameterValueType parameterValueType;
        private final Map<Class<? extends ConditionParameterDefinition>,
                ConditionParameterDefinition> augmentation;

        private ConditionParameterDefinitionImpl(
                ConditionParameterDefinitionBuilder base) {
            if (base.key == null) {
                key = new ConditionParameterDefinitionKey(base.parameterName);
                parameterName = base.parameterName;
            } else {
                key = base.key;
                parameterName = key.getParameterName();
            }
            parameterMatchPatterns = base.parameterMatchPatterns;
            parameterValueType = base.parameterValueType;
            augmentation = new HashMap<>(base.augmentation);
        }

        @Override
        public ConditionParameterDefinitionKey getKey() {
            return key;
        }

        @Override
        public ParameterMatchPatterns getParameterMatchPatterns() {
            return parameterMatchPatterns;
        }

        @Override
        public ParameterName getParameterName() {
            return parameterName;
        }

        @Override
        public ParameterValueType getParameterValueType() {
            return parameterValueType;
        }

        @Override
        public int hashCode() {
            return Objects.hash(key, parameterMatchPatterns,
                    parameterName, parameterValueType, augmentation);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if( obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            ConditionParameterDefinition other = (ConditionParameterDefinition) obj;
            if (!Objects.equals(key, other.getKey())) {
                return false;
            }
            if (!Objects.equals(parameterMatchPatterns, other.getParameterMatchPatterns())) {
                return false;
            }
            if (!Objects.equals(parameterName, other.getParameterName())) {
                return false;
            }
            if (!Objects.equals(parameterValueType, other.getParameterValueType())) {
                return false;
            }
            ConditionParameterDefinitionImpl otherImpl = (ConditionParameterDefinitionImpl) obj;
            return Objects.equals(augmentation, otherImpl.augmentation);
        }

        @Override
        public String toString() {
            return toStringHelper(this)
                    .add("key", key)
                    .add("parameterMatchPatterns", parameterMatchPatterns)
                    .add("parameterName", parameterName)
                    .add("parameterValueType", parameterValueType)
                    .add("augmentation", augmentation)
                    .toString();
        }
    }
}